@@ -16,8 +16,6 @@
 -- some operations. These bounds hold even in a persistent (shared) setting.
 --
 -- This implementation is based on a binomial heap augmented with a global root.
--- The spine of the heap is maintained lazily. To force the spine of the heap,
--- use 'seqSpine'.
 --
 -- This implementation does not guarantee stable behavior.
 --
@@ -76,6 +74,8 @@ module Data.PQueue.Max (
   mapU,
   foldrU,
   foldlU,
+  foldlU',
+  foldMapU,
   elemsU,
   toListU,
   -- * Miscellaneous operations
@@ -87,14 +87,16 @@ import Control.DeepSeq (NFData(rnf))
 import Data.Maybe (fromMaybe)
 
 #if MIN_VERSION_base(4,9,0)
-import Data.Semigroup (Semigroup((<>)))
+import Data.Semigroup (Semigroup(..), stimesMonoid)
 #endif
 
+import Data.Foldable (foldl')
+
 import qualified Data.PQueue.Min as Min
 import qualified Data.PQueue.Prio.Max.Internals as Prio
-import Data.PQueue.Prio.Max.Internals (Down(..))
+import Data.PQueue.Internals.Down (Down(..))
 
-import Prelude hiding (null, take, drop, takeWhile, dropWhile, splitAt, span, break, (!!), filter)
+import Prelude hiding (null, map, take, drop, takeWhile, dropWhile, splitAt, span, break, (!!), filter)
 
 #ifdef __GLASGOW_HASKELL__
 import GHC.Exts (build)
@@ -110,7 +112,7 @@ build f = f (:) []
 -- Implemented as a wrapper around 'Min.MinQueue'.
 newtype MaxQueue a = MaxQ (Min.MinQueue (Down a))
 # if __GLASGOW_HASKELL__
-  deriving (Eq, Ord, Data, Typeable)
+  deriving (Eq, Ord, Data)
 # else
   deriving (Eq, Ord)
 # endif
@@ -140,60 +142,64 @@ instance Read a => Read (MaxQueue a) where
 #if MIN_VERSION_base(4,9,0)
 instance Ord a => Semigroup (MaxQueue a) where
   (<>) = union
+  stimes = stimesMonoid
 #endif
 
 instance Ord a => Monoid (MaxQueue a) where
   mempty = empty
+#if !MIN_VERSION_base(4,11,0)
   mappend = union
+#endif
+  mconcat = unions
 
--- | /O(1)/. The empty priority queue.
+-- | \(O(1)\). The empty priority queue.
 empty :: MaxQueue a
 empty = MaxQ Min.empty
 
--- | /O(1)/. Is this the empty priority queue?
+-- | \(O(1)\). Is this the empty priority queue?
 null :: MaxQueue a -> Bool
 null (MaxQ q) = Min.null q
 
--- | /O(1)/. The number of elements in the queue.
+-- | \(O(1)\). The number of elements in the queue.
 size :: MaxQueue a -> Int
 size (MaxQ q) = Min.size q
 
--- | /O(1)/. Returns the maximum element of the queue. Throws an error on an empty queue.
+-- | \(O(1)\). Returns the maximum element of the queue. Throws an error on an empty queue.
 findMax :: MaxQueue a -> a
 findMax = fromMaybe (error "Error: findMax called on empty queue") . getMax
 
--- | /O(1)/. The top (maximum) element of the queue, if there is one.
+-- | \(O(1)\). The top (maximum) element of the queue, if there is one.
 getMax :: MaxQueue a -> Maybe a
 getMax (MaxQ q) = unDown <$> Min.getMin q
 
--- | /O(log n)/. Deletes the maximum element of the queue. Does nothing on an empty queue.
+-- | \(O(\log n)\). Deletes the maximum element of the queue. Does nothing on an empty queue.
 deleteMax :: Ord a => MaxQueue a -> MaxQueue a
 deleteMax (MaxQ q) = MaxQ (Min.deleteMin q)
 
--- | /O(log n)/. Extracts the maximum element of the queue. Throws an error on an empty queue.
+-- | \(O(\log n)\). Extracts the maximum element of the queue. Throws an error on an empty queue.
 deleteFindMax :: Ord a => MaxQueue a -> (a, MaxQueue a)
 deleteFindMax = fromMaybe (error "Error: deleteFindMax called on empty queue") . maxView
 
--- | /O(log n)/. Extract the top (maximum) element of the sequence, if there is one.
+-- | \(O(\log n)\). Extract the top (maximum) element of the sequence, if there is one.
 maxView :: Ord a => MaxQueue a -> Maybe (a, MaxQueue a)
 maxView (MaxQ q) = case Min.minView q of
   Nothing -> Nothing
   Just (Down x, q')
           -> Just (x, MaxQ q')
 
--- | /O(log n)/. Delete the top (maximum) element of the sequence, if there is one.
+-- | \(O(\log n)\). Delete the top (maximum) element of the sequence, if there is one.
 delete :: Ord a => MaxQueue a -> Maybe (MaxQueue a)
 delete = fmap snd . maxView
 
--- | /O(1)/. Construct a priority queue with a single element.
+-- | \(O(1)\). Construct a priority queue with a single element.
 singleton :: a -> MaxQueue a
 singleton = MaxQ . Min.singleton . Down
 
--- | /O(1)/. Insert an element into the priority queue.
+-- | \(O(1)\). Insert an element into the priority queue.
 insert :: Ord a => a -> MaxQueue a -> MaxQueue a
 x `insert` MaxQ q = MaxQ (Down x `Min.insert` q)
 
--- | /O(log (min(n1,n2)))/. Take the union of two priority queues.
+-- | \(O(\log min(n_1,n_2))\). Take the union of two priority queues.
 union :: Ord a => MaxQueue a -> MaxQueue a -> MaxQueue a
 MaxQ q1 `union` MaxQ q2 = MaxQ (q1 `Min.union` q2)
 
@@ -201,29 +207,29 @@ MaxQ q1 `union` MaxQ q2 = MaxQ (q1 `Min.union` q2)
 unions :: Ord a => [MaxQueue a] -> MaxQueue a
 unions qs = MaxQ (Min.unions [q | MaxQ q <- qs])
 
--- | /O(k log n)/. Returns the @(k+1)@th largest element of the queue.
+-- | \(O(k \log n)\)/. Returns the @(k+1)@th largest element of the queue.
 (!!) :: Ord a => MaxQueue a -> Int -> a
 MaxQ q !! n = unDown ((Min.!!) q n)
 
 {-# INLINE take #-}
--- | /O(k log n)/. Returns the list of the @k@ largest elements of the queue, in descending order, or
+-- | \(O(k \log n)\)/. Returns the list of the @k@ largest elements of the queue, in descending order, or
 -- all elements of the queue, if @k >= n@.
 take :: Ord a => Int -> MaxQueue a -> [a]
 take k (MaxQ q) = [a | Down a <- Min.take k q]
 
--- | /O(k log n)/. Returns the queue with the @k@ largest elements deleted, or the empty queue if @k >= n@.
+-- | \(O(k \log n)\)/. Returns the queue with the @k@ largest elements deleted, or the empty queue if @k >= n@.
 drop :: Ord a => Int -> MaxQueue a -> MaxQueue a
 drop k (MaxQ q) = MaxQ (Min.drop k q)
 
--- | /O(k log n)/. Equivalent to @(take k queue, drop k queue)@.
+-- | \(O(k \log n)\)/. Equivalent to @(take k queue, drop k queue)@.
 splitAt :: Ord a => Int -> MaxQueue a -> ([a], MaxQueue a)
-splitAt k (MaxQ q) = (map unDown xs, MaxQ q') where
+splitAt k (MaxQ q) = (fmap unDown xs, MaxQ q') where
   (xs, q') = Min.splitAt k q
 
 -- | 'takeWhile', applied to a predicate @p@ and a queue @queue@, returns the
 -- longest prefix (possibly empty) of @queue@ of elements that satisfy @p@.
 takeWhile :: Ord a => (a -> Bool) -> MaxQueue a -> [a]
-takeWhile p (MaxQ q) = map unDown (Min.takeWhile (p . unDown) q)
+takeWhile p (MaxQ q) = fmap unDown (Min.takeWhile (p . unDown) q)
 
 -- | 'dropWhile' @p queue@ returns the queue remaining after 'takeWhile' @p queue@.
 dropWhile :: Ord a => (a -> Bool) -> MaxQueue a -> MaxQueue a
@@ -234,7 +240,7 @@ dropWhile p (MaxQ q) = MaxQ (Min.dropWhile (p . unDown) q)
 -- satisfy @p@ and second element is the remainder of the queue.
 --
 span :: Ord a => (a -> Bool) -> MaxQueue a -> ([a], MaxQueue a)
-span p (MaxQ q) = (map unDown xs, MaxQ q') where
+span p (MaxQ q) = (fmap unDown xs, MaxQ q') where
   (xs, q') = Min.span (p . unDown) q
 
 -- | 'break', applied to a predicate @p@ and a queue @queue@, returns a tuple where
@@ -243,104 +249,129 @@ span p (MaxQ q) = (map unDown xs, MaxQ q') where
 break :: Ord a => (a -> Bool) -> MaxQueue a -> ([a], MaxQueue a)
 break p = span (not . p)
 
--- | /O(n)/. Returns a queue of those elements which satisfy the predicate.
+-- | \(O(n)\). Returns a queue of those elements which satisfy the predicate.
 filter :: Ord a => (a -> Bool) -> MaxQueue a -> MaxQueue a
 filter p (MaxQ q) = MaxQ (Min.filter (p . unDown) q)
 
--- | /O(n)/. Returns a pair of queues, where the left queue contains those elements that satisfy the predicate,
+-- | \(O(n)\). Returns a pair of queues, where the left queue contains those elements that satisfy the predicate,
 -- and the right queue contains those that do not.
 partition :: Ord a => (a -> Bool) -> MaxQueue a -> (MaxQueue a, MaxQueue a)
 partition p (MaxQ q) = (MaxQ q0, MaxQ q1)
   where  (q0, q1) = Min.partition (p . unDown) q
 
--- | /O(n)/. Maps a function over the elements of the queue, and collects the 'Just' values.
+-- | \(O(n)\). Maps a function over the elements of the queue, and collects the 'Just' values.
 mapMaybe :: Ord b => (a -> Maybe b) -> MaxQueue a -> MaxQueue b
 mapMaybe f (MaxQ q) = MaxQ (Min.mapMaybe (\(Down x) -> Down <$> f x) q)
 
--- | /O(n)/. Maps a function over the elements of the queue, and separates the 'Left' and 'Right' values.
+-- | \(O(n)\). Maps a function over the elements of the queue, and separates the 'Left' and 'Right' values.
 mapEither :: (Ord b, Ord c) => (a -> Either b c) -> MaxQueue a -> (MaxQueue b, MaxQueue c)
 mapEither f (MaxQ q) = (MaxQ q0, MaxQ q1)
   where  (q0, q1) = Min.mapEither (either (Left . Down) (Right . Down) . f . unDown) q
 
--- | /O(n)/. Assumes that the function it is given is monotonic, and applies this function to every element of the priority queue.
+-- | \(O(n)\). Creates a new priority queue containing the images of the elements of this queue.
+-- Equivalent to @'fromList' . 'Data.List.map' f . toList@.
+map :: Ord b => (a -> b) -> MaxQueue a -> MaxQueue b
+map f (MaxQ q) = MaxQ (Min.map (\(Down x) -> Down (f x)) q)
+
+-- | \(O(n)\). Assumes that the function it is given is monotonic, and applies this function to every element of the priority queue.
 -- /Does not check the precondition/.
 mapU :: (a -> b) -> MaxQueue a -> MaxQueue b
 mapU f (MaxQ q) = MaxQ (Min.mapU (\(Down a) -> Down (f a)) q)
 
--- | /O(n)/. Unordered right fold on a priority queue.
+-- | \(O(n)\). Unordered right fold on a priority queue.
 foldrU :: (a -> b -> b) -> b -> MaxQueue a -> b
 foldrU f z (MaxQ q) = Min.foldrU (flip (foldr f)) z q
 
--- | /O(n)/. Unordered left fold on a priority queue.
+-- | \(O(n)\). Unordered monoidal fold on a priority queue.
+--
+-- @since 1.4.2
+foldMapU :: Monoid m => (a -> m) -> MaxQueue a -> m
+foldMapU f (MaxQ q) = Min.foldMapU (f . unDown) q
+
+-- | \(O(n)\). Unordered left fold on a priority queue. This is rarely
+-- what you want; 'foldrU' and 'foldlU'' are more likely to perform
+-- well.
 foldlU :: (b -> a -> b) -> b -> MaxQueue a -> b
 foldlU f z (MaxQ q) = Min.foldlU (foldl f) z q
 
+-- | \(O(n)\). Unordered strict left fold on a priority queue.
+--
+-- @since 1.4.2
+foldlU' :: (b -> a -> b) -> b -> MaxQueue a -> b
+foldlU' f z (MaxQ q) = Min.foldlU' (foldl' f) z q
+
 {-# INLINE elemsU #-}
 -- | Equivalent to 'toListU'.
 elemsU :: MaxQueue a -> [a]
 elemsU = toListU
 
 {-# INLINE toListU #-}
--- | /O(n)/. Returns a list of the elements of the priority queue, in no particular order.
+-- | \(O(n)\). Returns a list of the elements of the priority queue, in no particular order.
 toListU :: MaxQueue a -> [a]
-toListU (MaxQ q) = map unDown (Min.toListU q)
+toListU (MaxQ q) = fmap unDown (Min.toListU q)
 
--- | /O(n log n)/. Performs a right-fold on the elements of a priority queue in ascending order.
+-- | \(O(n \log n)\). Performs a right-fold on the elements of a priority queue in ascending order.
 -- @'foldrAsc' f z q == 'foldlDesc' (flip f) z q@.
 foldrAsc :: Ord a => (a -> b -> b) -> b -> MaxQueue a -> b
 foldrAsc = foldlDesc . flip
 
--- | /O(n log n)/. Performs a left-fold on the elements of a priority queue in descending order.
+-- | \(O(n \log n)\). Performs a left-fold on the elements of a priority queue in descending order.
 -- @'foldlAsc' f z q == 'foldrDesc' (flip f) z q@.
 foldlAsc :: Ord a => (b -> a -> b) -> b -> MaxQueue a -> b
 foldlAsc = foldrDesc . flip
 
--- | /O(n log n)/. Performs a right-fold on the elements of a priority queue in descending order.
+-- | \(O(n \log n)\). Performs a right-fold on the elements of a priority queue in descending order.
 foldrDesc :: Ord a => (a -> b -> b) -> b -> MaxQueue a -> b
 foldrDesc f z (MaxQ q) = Min.foldrAsc (flip (foldr f)) z q
 
--- | /O(n log n)/. Performs a left-fold on the elements of a priority queue in descending order.
+-- | \(O(n \log n)\). Performs a left-fold on the elements of a priority queue in descending order.
 foldlDesc :: Ord a => (b -> a -> b) -> b -> MaxQueue a -> b
 foldlDesc f z (MaxQ q) = Min.foldlAsc (foldl f) z q
 
 {-# INLINE toAscList #-}
--- | /O(n log n)/. Extracts the elements of the priority queue in ascending order.
+-- | \(O(n \log n)\). Extracts the elements of the priority queue in ascending order.
 toAscList :: Ord a => MaxQueue a -> [a]
 toAscList q = build (\c nil -> foldrAsc c nil q)
 -- I can see no particular reason this does not simply forward to Min.toDescList. (lsp, 2016)
 
 {-# INLINE toDescList #-}
--- | /O(n log n)/. Extracts the elements of the priority queue in descending order.
+-- | \(O(n \log n)\). Extracts the elements of the priority queue in descending order.
 toDescList :: Ord a => MaxQueue a -> [a]
 toDescList q = build (\c nil -> foldrDesc c nil q)
 -- I can see no particular reason this does not simply forward to Min.toAscList. (lsp, 2016)
 
 {-# INLINE toList #-}
--- | /O(n log n)/. Returns the elements of the priority queue in ascending order. Equivalent to 'toDescList'.
+-- | \(O(n \log n)\). Returns the elements of the priority queue in ascending order. Equivalent to 'toDescList'.
 --
 -- If the order of the elements is irrelevant, consider using 'toListU'.
 toList :: Ord a => MaxQueue a -> [a]
-toList (MaxQ q) = map unDown (Min.toList q)
+toList (MaxQ q) = fmap unDown (Min.toList q)
 
 {-# INLINE fromAscList #-}
--- | /O(n)/. Constructs a priority queue from an ascending list. /Warning/: Does not check the precondition.
+-- | \(O(n)\). Constructs a priority queue from an ascending list. /Warning/: Does not check the precondition.
 fromAscList :: [a] -> MaxQueue a
-fromAscList = MaxQ . Min.fromDescList . map Down
+fromAscList = MaxQ . Min.fromDescList . fmap Down
 
 {-# INLINE fromDescList #-}
--- | /O(n)/. Constructs a priority queue from a descending list. /Warning/: Does not check the precondition.
+-- | \(O(n)\). Constructs a priority queue from a descending list. /Warning/: Does not check the precondition.
 fromDescList :: [a] -> MaxQueue a
-fromDescList = MaxQ . Min.fromAscList . map Down
+fromDescList = MaxQ . Min.fromAscList . fmap Down
 
 {-# INLINE fromList #-}
--- | /O(n log n)/. Constructs a priority queue from an unordered list.
+-- | \(O(n \log n)\). Constructs a priority queue from an unordered list.
 fromList :: Ord a => [a] -> MaxQueue a
-fromList = foldr insert empty
+fromList = MaxQ . Min.fromList . fmap Down
 
--- | /O(n)/. Constructs a priority queue from the keys of a 'Prio.MaxPQueue'.
+-- | \(O(n)\). Constructs a priority queue from the keys of a 'Prio.MaxPQueue'.
 keysQueue :: Prio.MaxPQueue k a -> MaxQueue k
 keysQueue (Prio.MaxPQ q) = MaxQ (Min.keysQueue q)
 
--- | /O(log n)/. Forces the spine of the heap.
+-- | \(O(\log n)\). @seqSpine q r@ forces the spine of @q@ and returns @r@.
+--
+-- Note: The spine of a 'MaxQueue' is stored somewhat lazily. Most operations
+-- take great care to prevent chains of thunks from accumulating along the
+-- spine to the detriment of performance. However, 'mapU' can leave expensive
+-- thunks in the structure and repeated applications of that function can
+-- create thunk chains.
 seqSpine :: MaxQueue a -> b -> b
 seqSpine (MaxQ q) = Min.seqSpine q
